@@ -765,6 +765,23 @@ attribute_is_sensitive (const CK_ATTRIBUTE *attr,
 	X (CKA_TRUST_STEP_UP_APPROVED)
 	X (CKA_CERT_SHA1_HASH)
 	X (CKA_CERT_MD5_HASH)
+	X (CKA_IBM_OPAQUE)
+	X (CKA_IBM_RESTRICTABLE)
+	X (CKA_IBM_NEVER_MODIFIABLE)
+	X (CKA_IBM_RETAINKEY)
+	X (CKA_IBM_ATTRBOUND)
+	X (CKA_IBM_KEYTYPE)
+	X (CKA_IBM_CV)
+	X (CKA_IBM_MACKEY)
+	X (CKA_IBM_USE_AS_DATA)
+	X (CKA_IBM_STRUCT_PARAMS)
+	X (CKA_IBM_STD_COMPLIANCE1)
+	X (CKA_IBM_PROTKEY_EXTRACTABLE)
+	X (CKA_IBM_PROTKEY_NEVER_EXTRACTABLE)
+	X (CKA_IBM_OPAQUE_PKEY)
+	X (CKA_IBM_DILITHIUM_KEYFORM)
+	X (CKA_IBM_DILITHIUM_RHO)
+	X (CKA_IBM_DILITHIUM_T1)
 	case CKA_VALUE:
 		return (klass != CKO_CERTIFICATE &&
 			klass != CKO_X_CERTIFICATE_EXTENSION);
