@@ -381,6 +381,27 @@ int        isatty           (int fd);
 
 #endif /* HAVE_ISATTY */
 
+#ifdef HAVE_READPASSPHRASE
+
+#include <readpassphrase.h>
+
+#else
+
+#define RPP_ECHO_OFF    0x00		/* Turn off echo (default). */
+#define RPP_ECHO_ON     0x01		/* Leave echo on. */
+#define RPP_REQUIRE_TTY 0x02		/* Fail if there is no tty. */
+#define RPP_FORCELOWER  0x04		/* Force input to lower case. */
+#define RPP_FORCEUPPER  0x08		/* Force input to upper case. */
+#define RPP_SEVENBIT    0x10		/* Strip the high bit from input. */
+#define RPP_STDIN       0x20		/* Read from stdin, not /dev/tty */
+
+char      *readpassphrase   (const char *prompt,
+                             char *buf,
+                             size_t bufsiz,
+                             int flags);
+
+#endif /* HAVE_READPASSPHRASE */
+
 /* If either locale_t or newlocale() is not available, strerror_l()
  * cannot be used */
 #if !defined(HAVE_LOCALE_T) || !defined(HAVE_NEWLOCALE)
@@ -394,11 +415,14 @@ int        isatty           (int fd);
 
 #endif
 
-void       p11_strerror_r    (int errnum,
-                              char *buf,
-                              size_t buflen);
+void       p11_strerror_r      (int errnum,
+                                char *buf,
+                                size_t buflen);
 
-int        p11_ascii_tolower (int c);
-int        p11_ascii_toupper (int c);
+int        p11_ascii_tolower   (int c);
+int        p11_ascii_toupper   (int c);
+
+bool       p11_ascii_strcaseeq (const char *s1,
+                                const char *s2);
 
 #endif /* __COMPAT_H__ */
