@@ -36,15 +36,24 @@
 
 #include "array.h"
 #include "attrs.h"
+#include "compat.h"
 #include "debug.h"
 #include "iter.h"
 #include "pin.h"
 #include "private.h"
 
 #include <assert.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#define _(x) dgettext(PACKAGE_NAME, x)
+#else
+#define _(x) (x)
+#endif
+
 typedef struct _Callback {
 	p11_kit_iter_callback func;
 	void *callback_data;
@@ -66,6 +75,8 @@ struct p11_kit_iter {
 	CK_ATTRIBUTE *match_attrs;
 	CK_SLOT_ID match_slot_id;
 	Callback *callbacks;
+	char *pin_value;
+	char *pin_source;
 
 	/* The input modules */
 	p11_array *modules;
@@ -104,6 +115,8 @@ struct p11_kit_iter {
 	unsigned int with_slots : 1;
 	unsigned int with_tokens : 1;
 	unsigned int with_objects : 1;
+	unsigned int with_login : 1;
+	unsigned int with_sessions : 1;
 };
 
 /**
@@ -111,6 +124,8 @@ struct p11_kit_iter {
  * @P11_KIT_ITER_KIND_MODULE: The iterator is pointing to a module.
  * @P11_KIT_ITER_KIND_SLOT: The iterator is pointing to a slot.
  * @P11_KIT_ITER_KIND_TOKEN: The iterator is pointing to a token.
+ * @P11_KIT_ITER_KIND_SESSION: The iterator is pointing to a token with an
+ *   active session.
  * @P11_KIT_ITER_KIND_OBJECT: The iterator is pointing to an object.
  * @P11_KIT_ITER_KIND_UNKNOWN: The iterator doesn't point to anything.
  *
@@ -126,6 +141,8 @@ struct p11_kit_iter {
  * @P11_KIT_ITER_WITH_MODULES: Stop at each module while iterating.
  * @P11_KIT_ITER_WITH_SLOTS: Stop at each slot while iterating.
  * @P11_KIT_ITER_WITH_TOKENS: Stop at each token while iterating.
+ * @P11_KIT_ITER_WITH_SESSIONS: Stop at each token while iterating (after
+ *   opening a session).
  * @P11_KIT_ITER_WITHOUT_OBJECTS: Ignore objects while iterating.
  *
  * Various flags controlling the behavior of the iterator.
@@ -167,7 +184,9 @@ p11_kit_iter_new (P11KitUri *uri,
 	iter->with_modules = !!(behavior & P11_KIT_ITER_WITH_MODULES);
 	iter->with_slots = !!(behavior & P11_KIT_ITER_WITH_SLOTS);
 	iter->with_tokens = !!(behavior & P11_KIT_ITER_WITH_TOKENS);
+	iter->with_sessions = !!(behavior & P11_KIT_ITER_WITH_SESSIONS);
 	iter->with_objects = !(behavior & P11_KIT_ITER_WITHOUT_OBJECTS);
+	iter->with_login = !!(behavior & P11_KIT_ITER_WITH_LOGIN);
 
 	p11_kit_iter_set_uri (iter, uri);
 	return iter;
@@ -199,6 +218,7 @@ p11_kit_iter_set_uri (P11KitIter *iter,
 	CK_SLOT_INFO *sinfo;
 	CK_INFO *minfo;
 	CK_ULONG count;
+	const char *pin_value;
 
 	return_if_fail (iter != NULL);
 
@@ -224,6 +244,21 @@ p11_kit_iter_set_uri (P11KitIter *iter,
 			tinfo = p11_kit_uri_get_token_info (uri);
 			if (tinfo != NULL)
 				memcpy (&iter->match_token, tinfo, sizeof (CK_TOKEN_INFO));
+
+			pin_value = p11_kit_uri_get_pin_value (uri);
+			if (pin_value != NULL)
+				iter->pin_value = strdup (pin_value);
+			else {
+				/* If the PIN is not immediately available
+				 * through pin-value, keep pin-source for later
+				 * retrieval.
+				 */
+				const char *pin_source;
+
+				pin_source = p11_kit_uri_get_pin_source (uri);
+				if (pin_source != NULL)
+					iter->pin_source = strdup (pin_source);
+			}
 		}
 	} else {
 		/* Match any module version number and slot ID */
@@ -508,6 +543,32 @@ call_all_filters (P11KitIter *iter,
 #define COROUTINE_RETURN(name,i,x) do { iter->name ## _state = i; return x; case i:; } while (0)
 #define COROUTINE_END(name) }
 
+static P11KitPin *
+request_pin (P11KitIter *iter)
+{
+	P11KitPin *pin;
+	char *pin_description;
+
+	if (iter->pin_value) {
+		return p11_kit_pin_new_for_string (iter->pin_value);
+	}
+
+	if (asprintf (&pin_description,
+		      _("PIN for %.*s"),
+		      (int) p11_kit_space_strlen (iter->token_info.label,
+						  sizeof (iter->token_info.label)),
+		      iter->token_info.label) < 0) {
+		return NULL;
+	}
+
+	pin = p11_kit_pin_request (iter->pin_source,
+				   NULL,
+				   pin_description,
+				   P11_KIT_PIN_FLAGS_USER_LOGIN);
+	free (pin_description);
+	return pin;
+}
+
 static CK_RV
 move_next_session (P11KitIter *iter)
 {
@@ -542,7 +603,7 @@ move_next_session (P11KitIter *iter)
 			COROUTINE_RETURN (move_next_session, 1, CKR_OK);
 		}
 
-		if (iter->with_slots || iter->with_tokens || iter->with_objects) {
+		if (iter->with_slots || iter->with_tokens || iter->with_sessions || iter->with_objects) {
 			CK_SLOT_ID *slots;
 
 			rv = (iter->module->C_GetSlotList) (CK_TRUE, NULL, &num_slots);
@@ -563,7 +624,8 @@ move_next_session (P11KitIter *iter)
 	}
 
 	/* Move to the next slot, and open a session on it */
-	while ((iter->with_slots || iter->with_tokens || iter->with_objects) &&
+	while ((iter->with_slots || iter->with_tokens || iter->with_sessions ||
+		iter->with_objects) &&
 	       iter->saw_slots < iter->num_slots) {
 		iter->slot = iter->slots[iter->saw_slots++];
 
@@ -580,6 +642,7 @@ move_next_session (P11KitIter *iter)
 		rv = (iter->module->C_GetTokenInfo) (iter->slot, &iter->token_info);
 		if (rv != CKR_OK || !p11_match_uri_token_info (&iter->match_token, &iter->token_info))
 			continue;
+
 		if (iter->with_tokens) {
 			iter->kind = P11_KIT_ITER_KIND_TOKEN;
 			COROUTINE_RETURN (move_next_session, 3, CKR_OK);
@@ -597,10 +660,31 @@ move_next_session (P11KitIter *iter)
 			return finish_iterating (iter, rv);
 
 		if (iter->session != 0) {
-			iter->move_next_session_state = 0;
-			iter->kind = P11_KIT_ITER_KIND_UNKNOWN;
-			return CKR_OK;
+			if (iter->with_login &&
+			    (iter->pin_value != NULL || iter->pin_source != NULL)) {
+				P11KitPin *pin;
+
+				pin = request_pin (iter);
+				if (!pin)
+					continue;
+
+				rv = (iter->module->C_Login) (iter->session, CKU_USER,
+							      (unsigned char *) p11_kit_pin_get_value (pin, NULL),
+							      p11_kit_pin_get_length (pin));
+				p11_kit_pin_unref (pin);
+				if (rv != CKR_OK)
+					return finish_iterating (iter, rv);
+			}
+
+			if (iter->with_sessions) {
+				iter->kind = P11_KIT_ITER_KIND_SESSION;
+				COROUTINE_RETURN (move_next_session, 4, CKR_OK);
+			}
 		}
+
+		iter->move_next_session_state = 0;
+		iter->kind = P11_KIT_ITER_KIND_UNKNOWN;
+		return CKR_OK;
 	}
 
 	COROUTINE_END (move_next_session);
@@ -644,7 +728,8 @@ p11_kit_iter_next (P11KitIter *iter)
 	if (iter->match_nothing)
 		return finish_iterating (iter, CKR_CANCEL);
 
-	if (!(iter->with_modules || iter->with_slots || iter->with_tokens || iter->with_objects))
+	if (!(iter->with_modules || iter->with_slots || iter->with_tokens ||
+	      iter->with_sessions || iter->with_objects))
 		return finish_iterating (iter, CKR_CANCEL);
 
 	/*
@@ -669,7 +754,8 @@ p11_kit_iter_next (P11KitIter *iter)
 	 * objects, or we are looking for modules/slots/tokens */
 	if ((iter->with_objects && iter->searched) ||
 	    (!iter->with_objects &&
-	     (iter->with_modules || iter->with_slots || iter->with_tokens))) {
+	     (iter->with_modules || iter->with_slots || iter->with_tokens ||
+	      iter->with_sessions))) {
 		/* Use iter->kind as the sentinel to detect the case where
 		 * any match (except object) is successful in
 		 * move_next_session() */
@@ -1081,6 +1167,8 @@ p11_kit_iter_free (P11KitIter *iter)
 	p11_attrs_free (iter->match_attrs);
 	free (iter->objects);
 	free (iter->slots);
+	free (iter->pin_value);
+	free (iter->pin_source);
 
 	for (cb = iter->callbacks; cb != NULL; cb = next) {
 		next = cb->next;
